local L = Capitalist.Locale
local realm = GetRealmName()
local character = UnitName("player")
local faction = UnitFactionGroup("player")

-- REVIEW: Duplicated from reagentUI.lua
local histogramColumns = 60
-- REVIEW: Duplicated from reagentUI.lua
local componentsPerPage = 5
-- REVIEW: Duplicated from reagentUI.lua
local widgetMinHeight = 0.01

function Capitalist:InitScanStats()
	self:DebugEntry("InitScanStats")

	self:DebugMisc(time())

	if (self.mode == "scan_component") then
		local name = self.ScanBuckets[1]
		--print("InitScanStats -- "..name)
		self.componentScanName = name
		self.componentScanData[name] = {}
		self.componentCurrentStack[name] = 1
		--self.componentScanCountTest = 0
		--self.componentScanCountAcceptableTest = 0
		--self.componentScanCountNoBuyoutTest = 0
		--self.componentScanCountOverpriceTest = 0
	end

	if (self.mode == "scan_products") then
	--	Capitalist.hit = 0
	--	Capitalist.miss = 0
		for productName, data in pairs(self.list) do
			data.ownAuctions = 0
			data.otherAuctions = {}
			data.cost = nil
			data.bestPrice = nil
			-- TODO: empty data.foundScanBucketIndex after each scan to save memory
			data.foundScanBucketIndex = nil
		end
	end

	self:DebugMisc(time())
--	self:DebugMisc(Capitalist.hit)
--	self:DebugMisc(Capitalist.miss)
	self:DebugExit("InitScanStats")
end

function Capitalist:EnsureProductCost(productName)
	local data = self.list[productName]
	if (not data.cost) then
		-- calculate cost of the product
		data.cost = self:CalculateSkillCopperCost(data.profession, data.skillName, false)

		-- add auction house cut to the price and minProfit
		data.cost = self.db.profile.minProfit + data.cost / 0.95
		-- TODO: math change - the formula above looks incorrect in two ways:
		-- 1. the auction house cut is 5% not 5.263% (i.e. 1/0.95)
		-- 2. if user specifies minprofit that should be net after AH cut
		-- data.cost = (self.db.profile.minProfit + data.cost) * 1.05
	end
end

function Capitalist:EnsureBestPrice(productName)
	self:EnsureProductCost(productName)
	local data = self.list[productName]
	if (not data.bestPrice) then
		local effectiveMargin = 0
		local effectiveMinMargin = 0
		local effectiveMaxMargin = 0
		if (not self.db.profile.optionsSpecific[realm][character].liquidate) then
			effectiveMargin = self.db.profile.margin
			effectiveMinMargin = self.db.profile.minMargin
			effectiveMaxMargin = self.db.profile.maxMargin
		end

		-- calculate maximum price
		data.bestPrice = data.cost * (1 + effectiveMargin / 100)
		if (data.bestPrice - data.cost < effectiveMinMargin) then
			data.bestPrice = data.cost + effectiveMinMargin
		end
		if (data.bestPrice - data.cost > effectiveMaxMargin) then
			data.bestPrice = data.cost + effectiveMaxMargin
		end
		data.bestPrice = math.floor(data.bestPrice) -- kills problems with float numbers
	end
end

function Capitalist:StartComponentScan(index)
	self:DebugEntry("StartComponentScan")

	if (self.updateHandler:IsVisible()) then
		self:Print("Cancelling previous "..self.mode)
		self.updateHandler:Hide()
	end

	for i = 1, histogramColumns do
		self.componentUI.items[index].histogram.segments[i]:SetColorTexture(i/histogramColumns, 1-(i/histogramColumns), 0)
	end

	local name = self.componentUI.items[index].name:GetText()
	local componentScanBuckets = {}
	table.insert(componentScanBuckets, name)

	self:StartScan("scan_component", componentScanBuckets, true)
	self:DebugExit("StartComponentScan")
end

function Capitalist:StartProductsScan()
	self:DebugEntry("StartProductsScan")

	if (self.updateHandler:IsVisible()) then
		self:Print("Cancelling previous "..self.mode)
		self.updateHandler:Hide()
	end

	self.auctionUI.lastScanLabel:Hide()
	self.auctionUI.progressLabel:Show()
	self.auctionUI.progressLabel:SetText(L["Scanning products"].."...")
	self.auctionUI.progress:Show()
	self.auctionUI.progress:SetWidth(1)
	self.auctionUI.scanButton:Disable()
	self.auctionUI.postQuotaButton:Disable()
	self.auctionUI.postAllButton:Disable()

	local productScanBuckets = self.productScanBuckets

	self:StartScan("scan_products", productScanBuckets, false)
	self:DebugExit("StartProductsScan")
end

function Capitalist:StartScan(scanMode, scanBuckets, exactMatch)
	self:DebugEntry("StartScan")

	if (scanBuckets and (#scanBuckets > 0)) then
		self.ScanBuckets = scanBuckets
		self.ScanBucketIndex = 1
		self.ScanBucketPage = 0
		self.ScanBucketCountBatch = nil
		self.ScanBucketCountTotal = nil
		self.ScanBucketCountRead = 0

		self.mode = scanMode

		self:InitScanStats()

		self:QueryBucket(exactMatch)
	
		self.updateHandler:Show()
	end

	self:DebugExit("StartScan")
end

function Capitalist:QueryBucket(exactMatch)
	self:DebugEntry("QueryBucket")
	self.ScanQueryPending = nil
	self.ScanAnalyzed = nil

	if (self.ScanBucketPage == 0) then
		if (self.mode == "scan_products") then
			self.auctionUI.progressLabel:SetText(string.format(L["Scanning products %d/%d"], self.ScanBucketIndex, #self.ScanBuckets))
		end
		self:DebugMisc(self.ScanBuckets[self.ScanBucketIndex])
	end

	self:DebugMisc(self.ScanBuckets[self.ScanBucketIndex]..", page "..self.ScanBucketPage)
	--print("Capitalist:QueryBucket - QAI --- "..self.ScanBuckets[self.ScanBucketIndex].."@"..self.ScanBucketPage)
	--QueryAuctionItems(self.productScanBuckets[self.ScanBucketIndex],nil,nil,0,0,0,self.ScanBucketPage,0,0,nil)
	--QueryAuctionItems(self.productScanBuckets[self.ScanBucketIndex],nil,nil,0,0,0,self.ScanBucketPage,0,0,false)
	--print("QAI")
	QueryAuctionItems(self.ScanBuckets[self.ScanBucketIndex],nil,nil,self.ScanBucketPage,nil,nil,false,exactMatch,nil)
	--print("/QAI")
	self:DebugExit("QueryBucket")
end

function Capitalist:OnUpdateScanning()
	--self:DebugEntry("OnUpdateScanning")
	--print("OnUpdateScanning")

	if (not self.InAuctionItemListUpdate) then
		--print("OnUpdateScanning B")
		if (self.ScanQueryPending) then
			--print("OnUpdateScanning C")
			if (CanSendAuctionQuery()) then
				--print("OnUpdateScanning D")
				self:QueryBucket()
			end
		end
	end

	--self:DebugExit("OnUpdateScanning")
end

function Capitalist:OnAuctionItemListUpdate()
	self:DebugEntry("OnAuctionItemListUpdate")
	--print("OnAuctionItemListUpdate")
	self.InAuctionItemListUpdate = true

	self.ScanBucketCountBatch, self.ScanBucketCountTotal = GetNumAuctionItems("list")
	self:AnalyzeAuctionPage()
	self.ScanBucketCountRead = self.ScanBucketCountRead + self.ScanBucketCountBatch

	self:DebugMisc("T "..self.ScanBucketCountTotal.." B "..self.ScanBucketCountBatch.." R "..self.ScanBucketCountRead)

	if ((self.ScanBucketCountBatch > 0) and
		(self.ScanBucketCountRead < self.ScanBucketCountTotal)) then

		self.ScanBucketPage = self.ScanBucketPage + 1
		self.ScanQueryPending = true
	else
		self.ScanBucketCountRead = 0
		self.ScanBucketIndex = self.ScanBucketIndex + 1
		if (self.ScanBucketIndex > #self.ScanBuckets) then
			self:ScanFinished()
		else
			self.ScanBucketPage = 0
			self.ScanQueryPending = true
		end
	end

	if (self.mode == "scan_products") then
		self.auctionUI.progress:SetWidth(200 * self.ScanBucketIndex / #self.ScanBuckets)
	end

	self.InAuctionItemListUpdate = false
	self:DebugExit("OnAuctionItemListUpdate")
end

function Capitalist:AnalyzeAuctionPage()
	self:DebugEntry("AnalyzeAuctionPage")
	
	self.ScanAnalyzed = true

	if (true) then
--		return
	end

	local batch, count = GetNumAuctionItems("list")
	self:DebugMisc(batch.."<batch count>"..count)

	local copperPrice = self:GetItemCopperShoppingPrice(self.componentScanName)
	local copperPriceMax = copperPrice * count * 4;
	if (self:HasResourceCopperCost(self.componentScanName)) then
		copperPriceMax = math.max(copperPriceMax, self:GetResourceCopperCost(self.componentScanName) * count * 4)
	end

	for i = 1, self.ScanBucketCountBatch do
		local name, _, count, _, _, _, _, _, _, buyoutPrice, _, _, _, owner, ownerFullName  = GetAuctionItemInfo("list", i)
--print(owner)
--print(ownerFullName)

		if (self.mode == "scan_component") then
			--self.componentScanCountTest = self.componentScanCountTest + 1

			-- REVIEW: This should always be true since we specify exactMatch in StartComponentScan
			--if (name == self.componentScanName) then
				if (buyoutPrice > 0) then
					-- REVIEW: Putting a trace here showed this getting called an absurd number of times while scanning a reagent.
					--local copperPrice = self:GetItemCopperShoppingPrice(name)

					--local copperPriceMax = copperPrice * count * 4;
					--if (self:HasResourceCopperCost(name)) then
					--	copperPriceMax = math.max(copperPriceMax, self:GetResourceCopperCost(name) * count * 4)
					--end

					-- REVIEW: Capture this criteria in a helper function and consider providing settings for user control/tuning.
					if ((copperPrice == 0) or
						(buyoutPrice < copperPriceMax)) then
						--self.componentScanCountAcceptableTest = self.componentScanCountAcceptableTest + 1
						--print("AnalyzeAuctionPage -- "..name)
						table.insert(self.componentScanData[name], {count = count, price = buyoutPrice / count, order = i + (self.ScanBucketPage - 1) * 50})
					else
						--self.componentScanCountOverpriceTest = self.componentScanCountOverpriceTest + 1
					end
				else
					--self.componentScanCountNoBuyoutTest = self.componentScanCountNoBuyoutTest + 1
				end
			--end

			-- REVIEW: Double-check QueryAuctionItems (and related) to see if there's a way to get the results sorted so that we can stop after the price is above the 60th column price.
			-- REVIEW: Double-check QueryAuctionItems (and related) to see if there's a way to get the results sorted such that the SearchStack "re-search completely" model can be improved.
		end

		if (self.mode == "scan_products") then
			-- REVIEW: round up instead of down?
			local buyoutUnitPrice = buyoutPrice / count

			-- TODO: Add concept of "my" w/r/t "same account" auctions

			local auctionLink = GetAuctionItemLink("list", i)
			if (auctionLink) then
				name = self:ApplyGlyphProductNameCorrection(name, auctionLink)
			end

			local v = self.list[name]
			if (v) then
				if (self:ProductIsEnabled(v.profession, v.skillName)) then

					if (not v.foundScanBucketIndex) then
						v.foundScanBucketIndex = self.ScanBucketIndex
						self:EnsureBestPrice(name)
					end

					if (v.foundScanBucketIndex == self.ScanBucketIndex) then
						self:EnsureBestPrice(name)
						if (owner == character) then
							v.ownAuctions = v.ownAuctions + 1
							if (v.buyoutSelf == nil or buyoutUnitPrice < v.buyoutSelf) then
								v.buyoutSelf = buyoutUnitPrice
							end
						else
							if (buyoutUnitPrice > 0) then
								table.insert(v.otherAuctions, buyoutUnitPrice)
								if ((v.buyoutOthers == nil) or
									(buyoutUnitPrice < v.buyoutOthers)) then
									v.buyoutOthers = buyoutUnitPrice
									v.otherOwner = owner
								end
							end
						end
					else
	--					print("Excluding "..name)
	--					print("  old bucket:"..self.productScanBuckets[v.foundScanBucketIndex])
	--					print("  new bucket:"..self.productScanBuckets[self.ScanBucketIndex])
					end
				end
			end
		end
	end

	if (self.mode == "scan_component") then
		for i = 1, componentsPerPage do
			-- REVIEW: This is a bit round-about.  Shouldn't the index be the key and the name be a value fetched from that key.
			--         Or put another way...It seems inefficient to walk through all (5) items doing a string compare to find which histogram to update.	
			if (self.componentUI.items[i].name:GetText() == self.componentScanName) then
				local numBatchAuctions, totalAuctions = GetNumAuctionItems("list");
				local progress = math.floor(histogramColumns * self.ScanBucketPage * 50 / totalAuctions)
			
				-- TODO: This is inefficient.  It's updating all 60 (histogramColumns) for every iteration instead of just the ones uncovered.
				-- Better algorithm would clear them all to widgetMinHeight and bump the appropriate ones up to full height as it progresses.
				-- Probably not a huge issue as it stands, but might be significant if/when "auto-scanning" of components is implemented.

				for j = 1, progress do
					-- REVIEW: These hard-coded UI metrics (50) all over the places could use a home for their constants.
					--print(i)
					--print(j)
					--print(self.componentUI)
					--print(self.componentUI.items)
					--print(self.componentUI.items[i])
					--print(self.componentUI.items[i].histogram)
					--print(self.componentUI.items[i].histogram.segments)
					--print(#self.componentUI.items[i].histogram.segments)
					--REVIEW: This IF check wasn't required (long) before 7.0
					if (j < #self.componentUI.items[i].histogram.segments) then
					--	print(self.componentUI.items[i].histogram.segments[j])
						self.componentUI.items[i].histogram.segments[j]:SetHeight(50)
					end
				end
			
				for j = progress + 1, histogramColumns do
					--REVIEW: This IF check wasn't required (long) before 7.0
					--if (j < #self.componentUI.items[i].histogram.segments) then
						--print(self.componentUI.items[i].histogram.segments[j])
						self.componentUI.items[i].histogram.segments[j]:SetHeight(widgetMinHeight)
						--print("aftersegmentsprint")
					--end
				end
			end
		end
	end
	
	self:DebugExit("AnalyzeAuctionPage")
end

function Capitalist:ScanFinished()
	self:DebugEntry("ScanFinished")
	--print("ScanFinished")
	self.updateHandler:Hide()

	if (self.mode == "scan_component") then
		--print("self.componentScanCountTest = "..self.componentScanCountTest)
		--print("self.componentScanCountAcceptableTest = "..self.componentScanCountAcceptableTest)
		--print("self.componentScanCountNoBuyoutTest = "..self.componentScanCountNoBuyoutTest)
		--print("self.componentScanCountOverpriceTest = "..self.componentScanCountOverpriceTest)

		-- REVIEW: A setting to buy stacks biggest first vs. smallest first?
		table.sort(self.componentScanData[self.componentScanName], function(a, b)
														if (a.price == b.price) then
															return a.count > b.count
														else
															return a.price < b.price
														end
													end)

		-- REVIEW: This is a bit round-about.  Shouldn't the index be the key and the name be a value fetched from that key.
		--         Or put another way...It seems inefficient to walk through all (5) items doing a string compare to find which histogram to update.	
		for i = 1, componentsPerPage do
			if (self.componentUI.items[i].name:GetText() == self.componentScanName) then
				self:UpdateHistogram(i)
			end
		end
	end

	if (self.mode == "scan_products") then
		self.auctionUI.lastScanLabel:SetText(L["Last Scan: "]..date("%H:%M"))
		self.auctionUI.lastScanLabel:Show()
		self.auctionUI.progress:Hide()
		self.auctionUI.progressLabel:Hide()
		self.auctionUI.scanButton:Enable()
	
		-- consistency check
		for i = 1, GetNumAuctionItems("owner") do
			local name, _, _, _, _, _, _, _, _, _, _, _, _, _, _, saleStatus = GetAuctionItemInfo("owner", i);

			local auctionLink = GetAuctionItemLink("owner", i)
			if (auctionLink) then
				name = self:ApplyGlyphProductNameCorrection(name, auctionLink)
			end

			if ((saleStatus == 0) and self.list[name]) then
				local v = self.list[name]
				if (self:ProductIsEnabled(v.profession, v.skillName)) then
					self.list[name].ownAuctions = self.list[name].ownAuctions - 1
				end
			end
		end
	
		for k, v in pairs(self.list) do
			if (v.ownAuctions ~= 0) then
				self:DebugMisc("FAILED SCAN CAUSE:"..k.." @ "..v.ownAuctions)
	print("FAILED SCAN CAUSE:"..k.." @ "..v.ownAuctions)

				self.auctionUI.lastScanLabel:SetText(L["Last Scan: failed"])
				self.validScan = false
			end
		end

		-- calculating optimal prices by undercutting competition
		for k, v in pairs(self.list) do
			if (self:ProductIsEnabled(v.profession, v.skillName)) then
				for j = 1, #v.otherAuctions do
					if (v.otherAuctions[j] <= v.bestPrice) then
						if (v.otherAuctions[j] > v.cost) then
							v.bestPrice = v.otherAuctions[j] - 1
						end
					end
				end
			end
		end

		self.validScan = true
		self:UpdateAuctionUI()
		self.auctionUI.postQuotaButton:Enable()
		self.auctionUI.postAllButton:Enable()
		self:CheckForCancel()
	end

	self:DebugExit("ScanFinished")
end

function Capitalist:CheckForCancel()
	self.toCancel = {}

	for i = 1, GetNumAuctionItems("owner") do
		local name, _, count, _, _, _, _, _, _, buyoutPrice, bidAmount = GetAuctionItemInfo("owner", i);

		-- REVIEW: round up instead of down?
		local buyoutUnitPrice = buyoutPrice / count

		local auctionLink = GetAuctionItemLink("owner", i)
		if (auctionLink) then
			name = self:ApplyGlyphProductNameCorrection(name, auctionLink)
		end

		if (self.list[name] and (bidAmount == 0)) then
			self:EnsureBestPrice(name)
			if (buyoutUnitPrice > self.list[name].bestPrice and self.db.profile.cancelYellow) then
				table.insert(self.toCancel, i)
			elseif (buyoutUnitPrice < self.list[name].bestPrice and self.db.profile.cancelBlue) then
				table.insert(self.toCancel, i)
			elseif (self.db.profile.cancelDuration and GetAuctionItemTimeLeft("owner", i) ~= 4) then
				table.insert(self.toCancel, i)
			end
		end
	end

	if (#self.toCancel > 0) then
		self.auctionUI.cancelButton:Enable()
	else
		self.auctionUI.cancelButton:Disable()
	end
end

function Capitalist:CancelAuctions()
	self:DebugEntry("CancelAuctions")
	for i = 1, table.getn(self.toCancel) do
		CancelAuctionWithPrompt(self.toCancel[i])
	end
	self.toCancel = {}
	self.auctionUI.cancelButton:Disable()
	self:DebugExit("CancelAuctions")
end
